# -*- coding: future_annotations -*-
import difflib
import logging
import sys
from typing import TYPE_CHECKING

from nbsafety.tracing.trace_events import TraceEvent
from nbsafety.tracing.trace_manager import TraceManager
from .utils import make_safety_fixture

if TYPE_CHECKING:
    from typing import List, Union
    from types import FrameType

logging.basicConfig(level=logging.INFO)


_RECORDED_EVENTS = []


def patched_emit_event_fixture():
    _RECORDED_EVENTS.clear()
    original_emit_event = TraceManager._emit_event

    def _patched_emit_event(self, evt: Union[TraceEvent, str], *args, **kwargs):
        event = TraceEvent(evt) if isinstance(evt, str) else evt
        frame: FrameType = kwargs.get('_frame', sys._getframe().f_back)
        kwargs['_frame'] = frame
        if frame.f_code.co_filename.startswith('<ipython-input'):
            if not (
                (event == TraceEvent.call and self.call_depth == 0) or
                (event == TraceEvent.return_ and self.call_depth == 1)
            ):
                _RECORDED_EVENTS.append(event)
        return original_emit_event(self, evt, *args, **kwargs)
    TraceManager._emit_event = _patched_emit_event
    yield
    TraceManager._emit_event = original_emit_event


# Reset dependency graph before each test
_safety_fixture, run_cell_ = make_safety_fixture(
    extra_fixture=patched_emit_event_fixture,
    # trace_messages_enabled=True,
)


_DIFFER = difflib.Differ()


def throw_and_print_diff_if_recorded_not_equal_to(actual: List[TraceEvent]) -> None:
    assert _RECORDED_EVENTS == actual, '\n'.join(_DIFFER.compare(_RECORDED_EVENTS, actual))
    _RECORDED_EVENTS.clear()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def test_recorded_events_simple():
    assert _RECORDED_EVENTS == []
    run_cell('logging.info("foo")')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_recorded_events_two_stmts():
    assert _RECORDED_EVENTS == []
    run_cell('x = [1, 2, 3]')
    run_cell('logging.info(x)')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.before_list_literal,
        *([TraceEvent.list_elt] * 3),
        TraceEvent.after_list_literal,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,

        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_nested_chains_no_call():
    assert _RECORDED_EVENTS == []
    run_cell('logging.info("foo is %s", logging.info("foo"))')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.argument,

        # next events correspond to `logging.info("foo")`
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.argument,

        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_list_nested_in_dict():
    assert _RECORDED_EVENTS == []
    run_cell('x = {1: [2, 3, 4]}')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.before_dict_literal,

        TraceEvent.dict_key,
        TraceEvent.before_list_literal,
        *([TraceEvent.list_elt] * 3),
        TraceEvent.after_list_literal,
        TraceEvent.dict_value,

        TraceEvent.after_dict_literal,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_function_call():
    assert _RECORDED_EVENTS == []
    run_cell("""
def foo(x):
    return [x]
""")
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])
    run_cell('foo([42])')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.before_call,
        TraceEvent.before_list_literal,
        TraceEvent.list_elt,
        TraceEvent.after_list_literal,
        TraceEvent.argument,
        TraceEvent.call,
        TraceEvent.before_stmt,
        TraceEvent.before_return,
        TraceEvent.before_list_literal,
        TraceEvent.list_elt,
        TraceEvent.after_list_literal,
        TraceEvent.after_return,
        TraceEvent.return_,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_lambda_in_tuple():
    assert _RECORDED_EVENTS == []
    run_cell('x = (lambda: 42,)')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.before_tuple_literal,
        TraceEvent.before_lambda,
        TraceEvent.after_lambda,
        TraceEvent.tuple_elt,
        TraceEvent.after_tuple_literal,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_fancy_slices():
    assert _RECORDED_EVENTS == []
    run_cell("""
import numpy as np
class Foo:
    def __init__(self, x):
        self.x = x
foo = Foo(1)
arr = np.zeros((3, 3, 3))
""")
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,

        TraceEvent.before_stmt,
        TraceEvent.call,
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.return_,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,

        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.before_complex_symbol,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.call,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.after_assign_rhs,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.return_,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,

        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.before_tuple_literal,
        TraceEvent.tuple_elt,
        TraceEvent.tuple_elt,
        TraceEvent.tuple_elt,
        TraceEvent.after_tuple_literal,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])

    run_cell('logging.info(arr[foo.x:foo.x+1,...])')
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.before_call,
        TraceEvent.before_complex_symbol,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.after_complex_symbol,
        TraceEvent.before_complex_symbol,
        TraceEvent.attribute,
        TraceEvent.after_complex_symbol,
        TraceEvent.subscript_slice,
        TraceEvent.subscript,
        TraceEvent._load_saved_slice,
        TraceEvent.after_complex_symbol,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_for_loop():
    assert _RECORDED_EVENTS == []
    run_cell("""
for i in range(10):
    pass
""")
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,
    ] + [
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_loop_iter,
    # ] * 10 + [
    ] * 1 + [
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_while_loop():
    assert _RECORDED_EVENTS == []
    run_cell("""
i = 0
while i < 10:
    i += 1
""")
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
        TraceEvent.before_stmt,
    ] + [
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_loop_iter,
    # ] * 10 + [
    ] * 1 + [
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])


def test_for_loop_nested_in_while_loop():
    assert _RECORDED_EVENTS == []
    run_cell("""
i = 0
while i < 10:
    for j in range(2):
        i += 1
""")
    throw_and_print_diff_if_recorded_not_equal_to([
        TraceEvent.init_cell,
        TraceEvent.before_stmt,
        TraceEvent.before_assign_rhs,
        TraceEvent.after_assign_rhs,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
        TraceEvent.before_stmt,
    ] + [
        TraceEvent.before_stmt,
        TraceEvent.before_complex_symbol,
        TraceEvent.before_call,
        TraceEvent.argument,
        TraceEvent.after_call,
        TraceEvent.after_complex_symbol,

        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_loop_iter,
        # TraceEvent.before_stmt,
        # TraceEvent.after_stmt,
        # TraceEvent.after_loop_iter,

        TraceEvent.after_stmt,
        TraceEvent.after_loop_iter,
    # ] * 5 + [
    ] * 1 + [
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
    ])



# -*- coding: future_annotations -*-
import sys

# Remove the CWD from sys.path while we load stuff.
# This is added back by InteractiveShellApp.init_path()
# TODO: probably need to make this separate from nbsafety package so that we can
#  completely avoid imports until after removing cwd from sys.path
if sys.path[0] == '':
    del sys.path[0]

from ipykernel import kernelapp as app
from nbsafety.kernel import SafeKernel
app.launch_new_instance(kernel_class=SafeKernel)



# -*- coding: future_annotations -*-
import logging
from typing import TYPE_CHECKING, NamedTuple

from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    from typing import Iterable, Optional, Union

    # avoid circular imports
    from nbsafety.data_model.data_symbol import DataSymbol


logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class Timestamp(NamedTuple):
    cell_num: int
    stmt_num: int

    @classmethod
    def current(cls) -> Timestamp:
        return cls(nbs().cell_counter(), tracer().module_stmt_counter())

    @classmethod
    def uninitialized(cls) -> Timestamp:
        return cls(-1, -1)

    @property
    def is_initialized(self):
        return self > Timestamp.uninitialized()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Timestamp):
            raise TypeError("cannot compare non-timestamp value %s with timestamp %s" % (other, self))
        return tuple(self._asdict().values()) == tuple(other._asdict().values())

    @classmethod
    def update_usage_info(cls, symbols: Union[Optional[DataSymbol], Iterable[Optional[DataSymbol]]], exclude_ns=False):
        if symbols is None:
            return
        try:
            iter(symbols)  # type: ignore
        except TypeError:
            symbols = [symbols]  # type: ignore
        used_time = cls.current()
        for sym in symbols:  # type: ignore
            if sym is not None:
                sym.update_usage_info(used_time=used_time, exclude_ns=exclude_ns)



# -*- coding: future_annotations -*-
import logging
import sys

from nbsafety.singletons import nbs
from .utils import assert_bool, make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def stale_detected():
    return nbs().test_and_clear_detected_flag()


def assert_detected(msg=''):
    assert_bool(stale_detected(), msg=msg)


def assert_false_positive(msg=''):
    """
    Same as `assert_detected` but asserts a false positive.
    Helps with searchability of false positives in case we want to fix these later.
    """
    return assert_detected(msg=msg)


def assert_not_detected(msg=''):
    assert_bool(not stale_detected(), msg=msg)


def assert_false_negative(msg=''):
    """
    Same as `assert_not_detected` but asserts a false negative.
    Helps with searchability of false negatives in case we want to fix these later.
    """
    return assert_not_detected(msg=msg)


def test_simplest():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('a = 3')
    run_cell('logging.info(b)')
    assert_detected('should have detected b has stale dep on old a')


def test_readme_example():
    run_cell('def eval_model_1(): return 0.5')
    run_cell('def eval_model_2(): return 0.85')
    run_cell('models = {"model_1": eval_model_1, "model_2": eval_model_2}')
    output = """
best_acc, best_model = max((f(), name) for name, f in models.items())
logging.info(f'The best model was {best_model} with an accuracy of {best_acc}.')
    """
    run_cell(output)
    run_cell('def eval_model_1(): return 0.9')
    run_cell(output)
    assert_detected('`models` depends on stale `eval_model_1`')


# TODO: to get this working properly, post-call argument
#  symbols need to have dependencies on pre-call argument
#  symbols. The place to do this is in `DataSymbol.create_symbols_for_call_args(...)`
def test_passed_sym_captured_as_dep_for_mutated_obj():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
        
def mutate(foo, x):
    foo.x = x
""")
    run_cell('foo = Foo(5)')
    run_cell('y = 7')
    run_cell('mutate(foo, y)')
    run_cell('logging.info(foo.x)')
    assert_not_detected()
    run_cell('y = 42')
    run_cell('logging.info(foo.x)')
    assert_detected('`foo.x` depends on old value of `y`')


def test_subscript_dependency():
    run_cell('lst = [0, 1, 2]')
    run_cell('x = 5')
    run_cell('y = x + lst[0]')
    run_cell('lst[0] = 10')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `lst[0]`')


def test_long_chain():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('c = b + 1')
    run_cell('d = c + 1')
    run_cell('e = d + 1')
    run_cell('f = e + 1')
    assert_not_detected('everything OK so far')
    run_cell('a = 2')
    run_cell('logging.info(f)')
    assert_detected('f has stale dependency on old value of a')


def test_for_loop_liveness_check():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell('x = 42')
    run_cell("""
for i in range(2):
    logging.info(y)
    y = 7
""")
    assert_detected('first use of `y` in loop has stale dependency on old value of `x`')


def test_redef_after_stale_use():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('a = 3')
    run_cell("""
logging.info(b)
b = 7
""")
    assert_detected('b has stale dependency on old value of a')


def test_class_redef():
    classdef = """
class Foo:
    def __init__(self, x):
        self.x = x"""
    run_cell(classdef)
    run_cell('x = 5')
    run_cell('foo = Foo(x)')
    run_cell(classdef)
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of class `Foo`')


def test_subscript_dependency_fp():
    run_cell('lst = [0, 1, 2]')
    run_cell('x = 5')
    run_cell('y = x + lst[0]')
    run_cell('lst[1] = 10')
    run_cell('logging.info(y)')
    assert_not_detected('y depends only on unchanged lst[0] and not on changed lst[1]')


def test_comprehension_generator_vars_not_live():
    run_cell('x = 0')
    run_cell('y = x + 6')
    run_cell('x = 42')
    run_cell('lst = [y for y in range(1) for j in range(1)]')
    assert_not_detected('`y` is not live in the list comprehension')


def test_lambda_params_not_live():
    run_cell('x = 0')
    run_cell('y = x + 6')
    run_cell('x = 42')
    run_cell('f = lambda y: y * 2')
    assert_not_detected('`y` is not live in the lambda expr')


def test_lambda_arg():
    run_cell('x = 5')
    run_cell('y = 7')
    run_cell('lam = lambda x: x * 2')
    run_cell('z = lam(y + 3)')
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of updated `x`')
    run_cell('y = 43')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on old value of `y`')


def test_lambda_scope():
    run_cell("""
def foo(x):
    return lambda: x + 5
""")
    run_cell('x = 5')
    run_cell('y = 7')
    run_cell('lam = foo(y)')
    run_cell('z = lam()')
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of updated `x`')
    run_cell('y = 43')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on old value of `y`')


def test_lambda_with_kwarg_scope():
    run_cell('w = 99')
    run_cell("""
def foo(x):
    lam = lambda t=w: t + x + 5
    return lam
""")
    run_cell('x = 5')
    run_cell('y = 7')
    run_cell('lam = foo(y)')
    run_cell('z = lam()')
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of updated `x`')
    run_cell('w = 43')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on old value of `y`')


def test_fundef_params_not_live():
    run_cell('x = 0')
    run_cell('y = x + 6')
    run_cell('x = 42')
    run_cell("""
def f(y):
    return y + 9""")
    assert_not_detected('`y` is not live in the function')


def test_no_edge_for_comprehension_variable():
    run_cell('i = 0')
    run_cell('x = [i for i in range(10)]')
    run_cell('i = 1')
    run_cell('logging.info(x)')
    assert_not_detected('`x` does not depend on `i`')


# simple test about the basic assignment
def test_basic_assignment():
    run_cell('a = 1')
    run_cell('b = 2')
    run_cell('c = a+b')
    run_cell('d = c+1')
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    # redefine a here but not c and d
    run_cell('a = 7')
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_detected("Did not detect that c's reference was changed")

    run_cell('c = a+b')
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_detected("Did not detect that d's reference was changed")

    run_cell('d = c+1')
    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
    assert_not_detected("There should be no more dependency issue")


def test_empty_list_assignment():
    run_cell('a = [5]')
    run_cell('b = a + [6]')
    run_cell('logging.info(b)')
    run_cell('a = [6]')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `a`')
    run_cell('b = a + [7]')
    run_cell('a = []')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `a`')


# redefined function example from the project prompt
def test_redefined_function_in_list():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell("""
funcs_to_run = [foo,bar]
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    
    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    assert_detected("Did not detect that funcs_to_run's reference was changed")

    run_cell("""
funcs_to_run = [foo,bar]
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    assert_not_detected("There should be no more dependency issue")


# like before but the function is called in a list comprehension
def test_redefined_function_for_funcall_in_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = [foo(), bar()]')
    run_cell("""
accum = 0
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


# like before but we run the list through a function before iterating
def test_redefined_function_for_funcall_in_modified_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = tuple([foo(), bar()])')
    run_cell("""
accum = 0
# for ret in map(lambda x: x * 5, retvals):
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


def test_for_loop_with_map():
    run_cell("""
accum = 0
foo = [1, 2, 3, 4, 5]
for ret in map(lambda x: x * 5, foo):
    accum += ret
""")
    run_cell('logging.info(accum)')
    assert_not_detected('no stale dep foo -> accum')
    run_cell('foo = [0]')
    run_cell('logging.info(accum)')
    assert_detected('should detect stale dep foo -> accum unless only propagating past cell bounds')


def test_redefined_function_over_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7

def baz(lst):
    return map(lambda x: 3*x, lst)
""")
    run_cell('retvals = baz([foo(), bar()])')
    run_cell("""
accum = 0
for ret in map(lambda x: x * 5, retvals):
    accum += ret
""")
    run_cell("""
def baz(lst):
    return map(lambda x: 7*x, lst)
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `baz`')


# like before but the function is called in a tuple comprehension
def test_redefined_function_for_funcall_in_tuple_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = (foo(), bar())')
    run_cell("""
accum = 0
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


def test_symbol_callpoint():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell("""
def f():
    return y + 7
""")
    run_cell('z = f()')
    assert_not_detected()
    run_cell('x = 42')
    run_cell('z = f()')
    assert_detected('`y` as referenced in call of function `f` is stale')


def test_symbol_callpoint_2():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell("""
def f():
    y = 9
    return y + 7
""")
    run_cell('z = f()')
    assert_not_detected()
    run_cell('x = 42')
    run_cell('z = f()')
    assert_not_detected()


def test_symbol_callpoint_3():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell("""
def f():
    return y + 7
""")
    run_cell('z = f()')
    assert_not_detected()
    run_cell('x = 42')
    run_cell("""
def f():
    return y + 7
""")
    run_cell('z = f()')
    assert_detected('`y` as referenced in call of function `f` is stale')


def test_symbol_callpoint_4():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell("""
def f(y):
    return y + 7
""")
    run_cell('z = f(5)')
    assert_not_detected()
    run_cell('x = 42')
    run_cell('z = f(5)')
    assert_not_detected('`y` as referenced in call of function `f` is an arg and therefore not stale')


def test_function_arg_independent_of_outer_staleness():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell("""
def f(y):
    return y + 7
""")
    run_cell('z = f(5)')
    assert_not_detected()
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of `x`')


def test_lambda_capture():
    run_cell('x = 42')
    run_cell('lst = []')
    run_cell('lst.append(lambda elt: elt + x)')
    run_cell('y = lst[0](7)')
    run_cell('x = 43')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


# Tests about variables that have same name but in different scope.
# There shouldn't be any extra dependency because of the name.
def test_variable_scope():
    run_cell("""
def func():
    x = 6
""")
    run_cell('x = 7')
    run_cell('y = x')
    run_cell('z = func')
    run_cell('logging.info("%s, %s", y, z())')

    # change x inside of the function, but not x outside of the function
    run_cell('def func():\n    x = 10')
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Did not detect the dependency change in the function")

    run_cell('y = x')
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Updating y should not solve the dependency change inside of function func")

    run_cell('z = func')
    run_cell('logging.info("%s, %s", y, z())')
    assert_not_detected("Updating z should solve the problem")


def test_variable_scope_2():
    run_cell('def func():\n    x = 6')
    run_cell('x = 7')
    run_cell('y = x')
    run_cell('z = func')
    run_cell('logging.info("%s, %s", y, z())')

    # change x outside of the function, but not inside of the function
    run_cell('x = 10')
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Did not detect the dependency change outside of the function")

    run_cell('z = func')
    run_cell('logging.info("%s, %s", y, z())')
    assert_detected("Updating z should not solve the dependency change outside of function")

    run_cell('y = x')
    run_cell('logging.info("%s, %s", y, z())')
    assert_not_detected("Updating y should solve the problem")


def test_default_args():
    run_cell("""
x = 7
def foo(y=x):
    return y + 5
""")
    run_cell('a = foo()')
    assert_not_detected()
    run_cell('x = 10')
    assert_not_detected()
    run_cell('b = foo()')
    assert_detected("Should have detected stale dependency of fn foo() on x")


def test_same_pointer():
    # a and b are actually pointing to the same thing
    run_cell('a = [7]')
    run_cell('b = a')
    run_cell('c = b + [5]')

    run_cell('a[0] = 8')
    run_cell('logging.info(b)')
    assert_not_detected('`b` is an alias of `a`, updating a should automatically update `b` as well')
    run_cell('logging.info(c)')
    assert_detected('`c` does not point to the same thing as `a` or `b`, thus there is a stale dependency here')


def test_func_assign_objs():
    run_cell("""
a = [1]
b = [1]
c = [2]
d = [3]""")
    run_cell("""
def func(x, y=a):
    e = [c[0] + d[0]]
    f = [x[0] + y[0]]
    return f
""")
    run_cell('z = func(c)')
    run_cell('a = [4]')
    run_cell('logging.info(z[0])')
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell("""
def func(x, y=a):
    logging.info(b[0])
    e = [c[0] + d[0]]
    f = [x[0] + y[0]]
    %safety show_deps f[0]
    return f
# z = func(c)""")
    run_cell('z = func(c)')
    run_cell('%safety show_deps z[0]')
    run_cell('logging.info(z[0])')
    assert_not_detected()
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('c = [3]')
    run_cell('logging.info(z)')
    assert_detected("Should have detected stale dependency of z on c")
    run_cell('c = [77]')
    run_cell('logging.info(z[0])')
    assert_detected("Should have detected stale dependency of z on c")
    run_cell('z = func(c)')
    run_cell('logging.info(z[0])')
    assert_not_detected()
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('b = [40]')
    run_cell('d = [4]')
    run_cell('logging.info(z[0])')
    assert_not_detected("Changing b and d should not affect z")
    run_cell('logging.info(z)')
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_ints():
    run_cell("""
a = 1
b = 1
c = 2
d = 3
def func(x, y=a):
    e = c + d
    f = x + y
    return f
""")
    run_cell('z = func(c)')
    run_cell('a = 4')
    run_cell('logging.info(z)')
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell("""
def func(x, y=a):
    logging.info(b)
    e = c + d
    f = x + y
    return f
z = func(c)
""")
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('c = 3')
    run_cell('logging.info(z)')
    assert_detected("Should have detected stale dependency of z on c")
    run_cell('z = func(c)')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('b = 4')
    run_cell('d = 1')
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_helper_func():
    run_cell("""
x = 3
a = 4
def f():
    def g():
        logging.info(a)
        return x
    return g()
y = f()
""")
    run_cell('x = 4')
    run_cell('logging.info(y)')
    assert_detected("Should have detected stale dependency of y on x")
    run_cell('y = f()')
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell('a = 1')
    run_cell('logging.info(y)')
    assert_not_detected("Changing a should not affect y")


def test_func_assign_helper_func_2():
    run_cell("""
x = 3
a = 4
def f():
    def g():
        logging.info(a)
        return x
    return g
y = f()()
""")
    run_cell('x = 4')
    run_cell('logging.info(y)')
    assert_detected("Should have detected stale dependency of y on x")


def test_branching():
    run_cell('y = 7')
    run_cell('x = y + 3')
    run_cell("""
if True:
    b = 5
else:
    y = 7
""")
    run_cell('logging.info(x)')
    assert_not_detected('false positive on unchanged y')


def test_branching_2():
    run_cell('y = 7')
    run_cell('x = y + 3')
    run_cell("""
if False:
    b = 5
else:
    y = 9
""")
    run_cell('logging.info(x)')
    assert_detected('x depends on stale y')


def test_identity_checking():
    run_cell('y = 7')
    run_cell('x = y + 3')
    run_cell('y = 7')
    run_cell('logging.info(x)')
    assert_not_detected('`y` was not mutated')


def test_identity_checking_obj():
    # To get this working properly, we need to create datasyms for all of the values in the literal
    run_cell('y = [7]')
    run_cell('x = y + [3]')
    run_cell('y[0] = 7')
    run_cell('logging.info(x)')
    assert_not_detected('`y` was not mutated')


def test_identity_checking_obj_2():
    run_cell('x, y = [7], [8]')
    run_cell('z = y + [3]')
    run_cell('x[0] = 99')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of x')
    run_cell('y[0] = 8')
    run_cell('logging.info(z)')
    assert_not_detected('`y` was not mutated')
    run_cell('y[0] = 42')
    run_cell('logging.info(z)')
    assert_detected('`y` was mutated')


def test_identity_checking_obj_3():
    run_cell('d = {"y": 7}')
    run_cell('x = list(d.values()) + [3]')
    run_cell('d["y"] = 7')
    run_cell('logging.info(x)')
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"] = 8')
    run_cell('logging.info(x)')
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_4():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell('d = {"y": [7]}')
    run_cell('d["x"] = d["y"] + [3]')
    run_cell('d["y"][0] = 7')
    run_cell('logging.info(d["x"])')
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"][0] = 8')
    run_cell('logging.info(d["x"])')
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_5():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell('d = {"y": {0: 7}}')
    run_cell('d["x"] = {1:3, **d["y"]}')
    run_cell('d["y"][0] = 7')
    run_cell('logging.info(d["x"])')
    assert_not_detected('`d["y"]` was not mutated')
    run_cell('d["y"][0] = 8')
    run_cell('logging.info(d["x"])')
    assert_detected('`d["y"]` was mutated')


def test_identity_checking_obj_6():
    # To get this working properly, we need to create datasyms for literal namespaces recursively
    run_cell('lst = [[1, 2], 0]')
    run_cell('lst[1] = lst[0] + [3, 4]')
    run_cell('lst[0][1] = 2')
    run_cell('logging.info(lst[1])')
    assert_not_detected('`lst[0][1]` was not mutated')
    run_cell('lst[0][1] = 42')
    run_cell('logging.info(lst[1])')
    assert_detected('`lst[0][1]` was mutated')


def test_starred_assignment_rhs():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell('z = 2')
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell('a, b, c, d, e = [x + 1, y + 2, z + 3, *lst]')
    run_cell('z = 42')
    run_cell('logging.info(a)')
    assert_not_detected()
    run_cell('logging.info(b)')
    assert_not_detected()
    run_cell('logging.info(c)')
    assert_detected()
    run_cell('logging.info(d)')
    assert_not_detected()
    run_cell('logging.info(e)')
    assert_not_detected()
    run_cell('lst[0] = 7')
    run_cell('logging.info(d)')
    assert_detected()
    run_cell('logging.info(e)')
    assert_not_detected()
    run_cell('lst[1] = 8')
    run_cell('logging.info(e)')
    assert_detected()


def test_starred_assignment():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell('z = 2')
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell('s, *t = [x + 1, y + 2, z + 3, *lst]')
    run_cell('z = 42')
    run_cell('logging.info(s)')
    assert_not_detected()
    run_cell('logging.info(t[0])')
    assert_not_detected()
    run_cell('logging.info(t[1])')
    assert_detected()
    run_cell('x = 99')
    run_cell('logging.info(s)')
    assert_detected()
    run_cell('logging.info(t[0])')
    assert_not_detected()
    run_cell('y = 142')
    run_cell('logging.info(t[0])')
    assert_detected()


def test_starred_assignment_in_middle():
    run_cell('a, b, c, d, e = 1, 2, 3, 4, 5')
    run_cell('x, *star, y = [a, b, c, d, e]')
    run_cell('a += 1')
    run_cell('logging.info(x)')
    assert_detected()
    run_cell('logging.info(star)')
    assert_not_detected()
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell('e += 1')
    run_cell('logging.info(y)')
    assert_detected()
    run_cell('logging.info(star)')
    assert_not_detected()
    run_cell('c += 1')
    run_cell('logging.info(star)')
    assert_detected()


def test_attributes():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = x.x + 5')
    run_cell('x.x = 8')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale attrval x.x')


@skipif_known_failing
def test_attributes_2():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = x.x + 5')
    run_cell('%safety show_deps x')
    run_cell('%safety show_deps y')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale x')


def test_attribute_unpacking():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = Foo(6)')
    run_cell('w = 42')
    run_cell('z = 43')
    run_cell('x.x, y.x = w + 2, z + 3')
    run_cell('z = 9001')
    run_cell('logging.info(x.x)')
    assert_not_detected()
    run_cell('logging.info(x)')
    assert_not_detected()
    run_cell('logging.info(y.x)')
    assert_detected()
    run_cell('logging.info(y)')
    assert_detected()
    run_cell('y.x = z + 3')
    run_cell('logging.info(y.x)')
    assert_not_detected()
    run_cell('w = 99')
    run_cell('logging.info(x.x)')
    assert_detected()
    run_cell('logging.info(x)')
    assert_detected()
    run_cell('logging.info(y.x)')
    assert_not_detected()
    run_cell('logging.info(y)')
    assert_not_detected()


def test_attribute_unpacking_no_overwrite():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = Foo(6)')
    run_cell('w = 42')
    run_cell('z = 43')
    run_cell('x.x, y.x = w + 2, z + 3')
    run_cell('s, t = 12, 13')
    run_cell('x.x, y.x = x.x + s, y.x + t')
    run_cell('w = 101')
    run_cell('logging.info(x.x)')
    assert_detected()
    run_cell('logging.info(y.x)')
    assert_not_detected()
    run_cell('z = 103')
    run_cell('logging.info(y.x)')
    assert_detected()


def test_attributes_3():
    run_cell("""
class Foo:
    def __init__(self, x, y):
        self.x = x
        self.y = y
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo(7, 8)')
    run_cell('y = bar.x + 5')
    run_cell('foo.x = 8')
    run_cell('logging.info(y)')
    assert_not_detected('y does not depend on updated attrval foo.y')


def test_stale_use_of_attribute():
    run_cell("""
class Foo:
    def __init__(self, x, y):
        self.x = x
        self.y = y
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo(7, 8)')
    run_cell('foo.x = bar.x + bar.y')
    run_cell('bar.y = 42')
    run_cell('logging.info(foo.x)')
    assert_detected('`foo.x` depends on stale `bar.y`')


def test_attr_manager_active_scope_resets():
    run_cell("""
y = 10
class Foo:
    def f(self):
        y = 11
        return y
def f():
    return y
""")
    run_cell('foo = Foo()')
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell('x = foo.f() + f()')
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `y`')


def test_attribute_call_point():
    run_cell("""
y = 10
class Foo:
    def f(self):
        return y
""")
    run_cell('foo = Foo()')
    run_cell('x = foo.f()')
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `y`')


def test_attr_manager_active_scope_with_property():
    run_cell("""
y = 10
class Foo:
    @property
    def f(self):
        y = 11
        return y
""")
    run_cell('foo = Foo()')
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell('x = foo.f')
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of outer `y`')


def test_namespace_scope_resolution():
    run_cell("""
y = 42
class Foo:
    y = 10
    @property
    def foo(self):
        return y
""")
    run_cell('foo = Foo()')
    run_cell('x = foo.foo')
    run_cell('Foo.y = 99')
    run_cell('logging.info(x)')
    assert_not_detected('`x` should not have dependency on `Foo.y`')


def test_long_chain_attribute():
    run_cell("""
class Foo:
    shared = 99
    def __init__(self, x, y):
        self.x = x
        self.y = y + self.shared
        
    class Bar:
        def __init__(self, a, b):
            self.a = a
            self.b = b
            
        def foo(self):
            return Foo(self.a, self.b)
            
        def bar(self):
            return Foo
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo.Bar(7, 8)')
    run_cell('foo.x = 42')
    run_cell('logging.info(bar.a)')
    assert_not_detected()
    run_cell('Foo.Bar(9, 10).foo().shared = 100')
    run_cell("""
for _ in range(10**2):
    Foo.Bar(9, 10).foo().shared = 100
""")
    run_cell('logging.info(foo.y)')
    assert_not_detected('we mutated `shared` on obj and not on `Foo` (the one on which `foo.y` depends)')
    run_cell('Foo.Bar(9, 10).bar().shared = 100')
    run_cell('logging.info(foo.y)')
    assert_detected('we mutated a shared value on which `foo.y` depends')


def test_numpy_subscripting():
    run_cell('import numpy as np')
    run_cell('x = np.zeros(5)')
    run_cell('y = x[3] + 5')
    run_cell('x[3] = 2')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale x[3]')


def test_dict_subscripting():
    run_cell('d = {"foo": "bar", 0: "bat"}')
    run_cell('x = 7')
    run_cell('d[0] = x')
    run_cell('x += 1')
    run_cell('z = d["foo"] + " asdf"')
    assert_not_detected('`d["foo"]` does not depend on stale `x`, unlike `d[0]`')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('logging.info(d["foo"])')
    assert_not_detected()


def test_dict_subscripting_2():
    run_cell('d = {"foo": "bar", 0: "bat"}')
    run_cell('x = 7')
    run_cell('d[0] = x')
    run_cell('x += 1')
    run_cell('d[0] = x')
    run_cell('logging.info(d)')
    assert_not_detected('we updated the stale entry of `d`')


def test_pandas_attr_mutation_with_alias():
    run_cell('import pandas as pd')
    run_cell('df = pd.DataFrame({"a": [0,1], "b": [2., 3.]})')
    run_cell('asdf = df.b')
    run_cell('df.b *= 7')
    run_cell('logging.info(asdf)')
    assert_detected('`asdf` has same values as `df.b`, but they now point to different things, '
                    'which in general could be dangerous')


def test_stale_detection_works_when_namespace_available_but_stale_symbol_unavailable():
    run_cell('import pandas as pd')
    run_cell('data = {"a": list(range(5)), "b": list(range(1, 6))}')
    run_cell('df = pd.DataFrame(data)')
    run_cell('df.b')  # force creation of a namespace for df
    run_cell('data = {"a": list(range(4)), "b": list(range(1, 5))}')
    run_cell('logging.info(df.a)')
    assert_detected('`df.a` depends on stale `data` dictionary')


def test_stale_detection_works_when_namespace_available_but_stale_symbol_unavailable_2():
    run_cell('import pandas as pd')
    run_cell('data = {"a": list(range(5)), "b": list(range(1, 6))}')
    run_cell('df = pd.DataFrame(data)')
    run_cell('df.b')  # force creation of a namespace for df
    run_cell('data["b"] = 42')
    run_cell('logging.info(df.a)')
    assert_false_positive('`df.a` independent of entry `data["b"]`')


def test_list_alias_breaking():
    run_cell('x = [0]')
    run_cell('y = x')
    run_cell('x = [7]')
    run_cell('logging.info(y)')
    assert_detected('`y` has stale dependency on old `x`')


def test_list_alias_no_break():
    run_cell('x = [0]')
    run_cell('y = x')
    run_cell('x *= 7')
    run_cell('logging.info(y)')
    assert_not_detected('`y` still aliases `x`')


@skipif_known_failing
def test_subscript_sensitivity():
    run_cell('lst = list(range(5))')
    run_cell('i = 0')
    run_cell('lst[i] = 10')
    run_cell('i = 1')
    run_cell('logging.info(lst)')
    assert_detected('`lst[0]` depends on stale i')


def test_subscript_adds():
    run_cell('x = 42')
    run_cell("""
d = {
    'foo': x,
    'bar': 77
}
""")
    run_cell('d["bar"] = 99')
    run_cell('x = 100')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `x`')


def test_list_mutation():
    run_cell('lst = list(range(5))')
    run_cell('x = 42')
    run_cell('asdf = []')
    run_cell('asdf.append(lst.append(x))')
    run_cell('x = 43')
    run_cell('logging.info(lst)')
    assert_detected('lst depends on stale x')


def test_list_mutation_extend():
    run_cell('lst = [0, 1]')
    run_cell('x = lst[0] + 7')
    run_cell('lst.extend([2, 3, 4, x])')
    run_cell('logging.info(x)')
    assert_not_detected()
    run_cell('x = 77')
    run_cell('logging.info(lst)')
    assert_detected('`lst` depends on stale `x`')


def test_list_mutation_2():
    run_cell('lst = list(range(5))')
    run_cell('x = lst + [42, 43]')
    run_cell('lst.append(99)')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `lst`')


def test_list_mutation_from_attr():
    run_cell('s = "hello X world X how X are X you X today?"')
    run_cell('lst = []')
    run_cell("""
for word in s.split('X'):
    lst.append(word.strip())
""")
    run_cell('s = "foobar"')
    run_cell('logging.info(lst)')
    assert_detected('`lst` depends on stale `s`')


@skipif_known_failing
def test_list_mutation_extend_from_attr():
    # for this one, we somehow have to avoid disabling tracing which is hard
    run_cell('s = "hello X world X how X are X you X today?"')
    run_cell('lst = []')
    run_cell('lst.extend(word.strip() for word in s.split("X"))')
    run_cell('s = "foobar"')
    run_cell('logging.info(lst)')
    assert_detected('`lst` depends on stale `s`')


def test_lazy_class_scope_resolution():
    run_cell("""
class Foo:
    shared = 99
    def __init__(self, x):
        self.x = x
""")
    run_cell('foo = Foo(10)')
    run_cell('y = 11')
    run_cell('Foo.shared = y + 42')
    run_cell('y = 12')
    run_cell('logging.info(foo.shared)')
    assert_detected('`foo.shared` should point to same DataSymbol as `Foo.shared` and thus also has stale dep')
    run_cell('foo.shared = 89')
    run_cell('logging.info(Foo.shared)')
    assert_detected('Even though we refreshed `foo.shared`, this '
                    'has no bearing on the original class member `Foo.shared`')


def test_new_scope_val_depends_on_old():
    run_cell("""
class Foo:
    shared = 99
""")
    run_cell('foo = Foo()')
    run_cell('foo.shared = 11')
    run_cell('foo_shared_alias = foo.shared')
    run_cell('Foo.shared = 12')
    run_cell('logging.info(foo_shared_alias)')
    assert_detected()
    run_cell('logging.info(foo.shared)')
    assert_detected()
    run_cell('logging.info(foo)')
    assert_detected()


def test_class_member_mutation_does_not_affect_instance_members():
    run_cell("""
class Foo:
    shared = 99
    def __init__(self):
        self.x = 42
""")
    run_cell('foo = Foo()')
    run_cell('Foo.shared = 12')
    run_cell('logging.info(foo.x)')
    assert_not_detected()


def test_numpy_subscripting_fp():
    run_cell('import numpy as np')
    run_cell('x = np.zeros(5)')
    run_cell('y = x[3] + 5')
    run_cell('x[0] = 2')
    run_cell('logging.info(y)')
    assert_not_detected('`y` depends on unchanged `x[3]` and not on changed `x[0]`')


def test_old_format_string():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = "{} + {} = {}".format(a, b, a + b)')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_old_format_string_kwargs():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = "{a} + {b} = {total}".format(a=a, b=b, total=a + b)')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_new_format_string():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = f"{a} + {b} = {a+b}"')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_scope_resolution():
    run_cell("""
def f(x):
    def g(x):
        return 2 * x
    return g(x) + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_scope_resolution_2():
    run_cell("""
def g(x):
    return 2 * x
def f(x):
    return g(x) + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg():
    run_cell("""
def f(y):
    return 2 * y + 8
""")
    run_cell('x = 7')
    run_cell('z = f(y=x)')
    run_cell('x = 8')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on stale `x`')


def test_funcall_kwarg_2():
    run_cell("""
def f(y):
    return 2 * y + 8
""")
    run_cell('x = 7')
    run_cell('y = f(y=x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg_3():
    run_cell("""
def f(x):
    return 2 * x + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x=x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg_4():
    run_cell("""
def f(x):
    return 2 * x + 8
""")
    run_cell('x = 7')
    run_cell('x = f(x=x)')
    run_cell('x = 8')
    run_cell('logging.info(x)')
    assert_not_detected('`x` is overriden so should not be stale')


def test_single_line_dictionary_literal():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')


def test_single_line_dictionary_literal_fix_stale_deps():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')
    run_cell('d[foo] = bar')
    assert_not_detected()
    run_cell('logging.info(d)')
    # assert_false_positive('`d`s stale dep fixed, but this is hard to detect '
    #                       'since we did not yet have a DataSymbol for `d[foo]` when staleness introduced')
    assert_not_detected('`d`s stale dep fixed')
    run_cell('foo = 8')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `foo`')
    # TODO: not sure what the correct behavior for the below write to d[foo] after foo changed should be
    # run_cell('d[foo] = bar')
    # run_cell('logging.info(d)')
    # assert_not_detected('`d`s stale dep fixed')


def test_multiline_dictionary_literal():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell("""
d = {
    foo: bar,
    'pi': 42,
}
""")
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')


def test_exception():
    run_cell('lst = list(range(5))')
    run_cell('x = 6')
    run_cell("""
try:
    lst[x] = 42
except:
    lst[0] = 42
""")
    run_cell('x = 7')
    run_cell('logging.info(lst)')
    assert_not_detected('lst should be independent of x due to exception')


def test_for_loop_binding():
    run_cell('a = 0')
    run_cell('b = 1')
    run_cell('c = 2')
    run_cell('lst = [a, b, c]')
    run_cell("""
for i in lst:
    pass
""")
    run_cell('a = 3')
    run_cell('logging.info(i)')
    assert_false_positive('`i` should not depend on `a` at end of for loop but this is hard')


@skipif_known_failing
def test_for_loop_literal_binding():
    run_cell('a = 0')
    run_cell('b = 1')
    run_cell('c = 2')
    run_cell("""
for i in [a, b, c]:
    pass
""")
    run_cell('a = 3')
    run_cell('logging.info(i)')
    assert_not_detected('`i` should not depend on `a` at end of for loop')


def test_for_loop_partial_dep():
    run_cell('lst = list(range(10))')
    run_cell('s = 0')
    run_cell("""
for i in range(5):
    s += lst[i]
""")
    run_cell('lst[-1] = 42')
    run_cell('logging.info(s)')
    assert_not_detected('`s` does not depend on last entry of `lst`')
    run_cell('lst[1] = 22')
    run_cell('logging.info(s)')
    assert_false_negative('`s` does depend on second entry of `lst` but tracing every iteration of loop is slow')


def test_for_loop_tuple_unpack():
    run_cell('x = (1, 2)')
    run_cell('y = (3, 4)')
    run_cell('total_i = 0')
    run_cell('total_j = 0')
    run_cell("""
for i, j in [x, y]:
    total_i += i
    total_j += j
""")
    run_cell('logging.info(total_i)')
    assert_not_detected()
    run_cell('logging.info(total_j)')
    assert_not_detected()
    run_cell('x = (42, 43)')
    run_cell('logging.info(total_i)')
    assert_detected('`total_i` depends on old version of `x`')
    run_cell('logging.info(total_j)')
    assert_detected('`total_j` also depends on old version of `x`')


def test_same_cell_redefine():
    run_cell('a = 0')
    run_cell("""
b = a + 1
a = 42
""")
    run_cell('logging.info(b)')
    assert_not_detected('`b` should not be considered as having stale dependency since `a` changed in same cell as `b`')


def test_multiple_stmts_in_one_line():
    run_cell('a = 1; b = 2')
    run_cell('x = a + b')
    run_cell('a = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale value of `a`')


def test_multiple_stmts_in_one_line_2():
    run_cell('a = 1; b = 2')
    run_cell('x = a + b')
    run_cell('b = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale value of `a`')


def test_line_magic():
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
%lsmagic
%lsmagic
a = 0
""")
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
%lsmagic
x = a + 1
""")
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
a = 42
""")
    run_cell("""
%lsmagic
%lsmagic
logging.info(x)
""")
    assert_detected('`x` depends on stale value of `a`')
    run_cell("""
%lsmagic
logging.info(x)
%lsmagic
""")
    assert_detected('`x` depends on stale value of `a`')


def test_exception_stack_unwind():
    def assert_stack_size(size):
        len_call_stack = 'len(TraceManager.instance().call_stack)'
        return ', '.join([
            f'assert {len_call_stack} == {size}',
            f'"%d vs {size}" % {len_call_stack}',
        ])
    run_cell(f"""
import numpy as np
from nbsafety.singletons import TraceManager
{assert_stack_size(0)}
def f():
    {assert_stack_size(1)}
    def g():
        {assert_stack_size(2)}
        def h():
            {assert_stack_size(3)}
            return np.loadtxt('does-not-exist.txt')
        return h()
    try:
        return g()
    except:
        {assert_stack_size(1)}
f()
{assert_stack_size(0)}
""")


def test_throwing_statements_do_not_track_deps():
    run_cell("""
z = 10
def foo():
    def bar():
        raise ValueError('foo!')
    return bar() + z
x = 0
y = x + 1
""")
    run_cell("""
try:
    x = 42 + foo()
except:
    pass
""")
    run_cell('logging.info(y)')
    assert_not_detected('no stale dep for `y` because update on `x` threw exception')
    run_cell('z = 99')
    run_cell('logging.info(x)')
    assert_not_detected('no stale dep for `x` because it is indep of `z` (attempted dep add threw)')


def test_attr_dep_from_somewhere_else():
    run_cell('import fakelib as fake')
    run_cell('fake.y = 7')
    run_cell('x = fake.y + 1')
    run_cell('fake.y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on old value of `fake.y`')


def test_attr_use_from_somewhere_else():
    run_cell('import fakelib as fake')
    run_cell('x = 7')
    run_cell('fake.y = x + 1')
    run_cell('x = 42')
    run_cell('logging.info(fake.y)')
    assert_detected('`fake.y` depends on old value of `x`')


def test_class_assignment():
    run_cell("""
class Foo:
    def __init__(self):
        self.y = 99

Bar = Foo
foo = Bar()
x = 7
""")
    run_cell('foo.y = x + 1')
    run_cell('x = 42')
    run_cell('logging.info(foo.y)')
    assert_detected('`foo.y` depends on stale `x`')
    run_cell('foo.y = 10')
    run_cell('x = foo.y + 1')
    run_cell('foo.y = 12')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `foo.y`')


def test_no_class_false_positives():
    run_cell('x = 7')
    run_cell('y = x + 1')
    run_cell('x = 42')
    run_cell("""
try:
    class Foo:
        logging.info(y)
except:
    pass
""")
    assert_not_detected('x inside class scope is different')


def test_set_literal():
    run_cell('x, y, z = 1, 2, 3')
    run_cell('s = {x + 1, y + 7}')
    run_cell('z = 42')
    run_cell('logging.info(s)')
    assert_not_detected()
    run_cell('x = 17')
    run_cell('logging.info(s)')
    assert_detected()


def test_tuple_unpack_simple():
    run_cell('x, y = 0, 1')
    run_cell('a, b = x + 2, y + 3')
    run_cell('x, y = 42, 43')
    run_cell('logging.info(a)')
    assert_detected('`a` depends on stale `x`')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `y`')


def test_tuple_unpack_hard():
    run_cell('x, y = 0, 1')
    run_cell('a, b = x + 2, y + 3')
    run_cell('y = 43')
    run_cell('logging.info(a)')
    assert_not_detected('`a` does not depend on `y`')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `y`')
    run_cell('b = y + 10')
    run_cell('x = 99')
    run_cell('logging.info(b)')
    assert_not_detected('`b` does not depend on `x`')
    run_cell('logging.info(a)')
    assert_detected('`a` depends on stale `x`')


def test_unpack_multiple_from_single():
    run_cell('x, y = 0, 1')
    run_cell('lst = [x + 1, y + 1]')
    run_cell('a, b = lst')
    run_cell('x = 42')
    run_cell('logging.info(b)')
    assert_not_detected()
    run_cell('logging.info(a)')
    assert_detected()


@skipif_known_failing
def test_attr_dep_with_top_level_overwrite():
    run_cell("""
class Foo:
    def __init__(self):
        self.y = 99
foo = Foo()
x = 42
foo.y = x + 7
""")
    run_cell('x = 43')
    run_cell('logging.info(foo)')  # this should be a 'deep' usage of foo
    assert_detected('logging.info could display `foo.y` which depends on x')
    run_cell('foo.y = 70')
    assert_not_detected('we just fixed stale dep of `foo.y` by changing its deps')
    run_cell('x = foo.y + 7')
    run_cell('foo = 81')
    run_cell('logging.info(x)')
    assert_detected('`x` has stale dep on `foo` (transitively through `foo.y`)')


def test_typed_assignment():
    run_cell('a = 0')
    run_cell('b: int = a + 1')
    run_cell('a: int = 42')
    run_cell('logging.info(b)')
    assert_detected('`b` has stale dep on `a`')


@skipif_known_failing
def test_time_line_magic():
    run_cell('a = 0')
    run_cell('b = a + 1')
    run_cell('%time a = 42')
    run_cell('logging.info(b)')
    assert_detected('`b` has stale dep on `a`')


@skipif_known_failing
def test_cell_magic():
    run_cell('a = 0')
    run_cell('b = a + 1')
    run_cell('%%time\na = 42')
    run_cell('logging.info(b)')
    assert_detected('`b` has stale dep on `a`')


def test_pandas():
    run_cell('import numpy as np')
    run_cell('import pandas as pd')
    run_cell('arr = 1 + np.arange(10)')
    run_cell('df = pd.DataFrame({"col": arr})')
    run_cell('df2 = df.dropna()')
    run_cell('df.dropna()')
    run_cell('logging.info(df2)')
    assert_not_detected('`df.dropna()` did not mutate `df`')
    run_cell('df.dropna(inplace=True)')
    run_cell('logging.info(df2)')
    assert_detected('`df.dropna(inplace=True)` mutated `df`')


def test_deeply_nested_arg_and_kwarg_refs_with_attr_calls():
    run_cell("""
class Foo:
    def foo(self, x, y=0):
        return x + y
""")
    run_cell('x = 1')
    run_cell('y = 2')
    run_cell('foo = Foo()')
    run_cell('z = foo.foo(foo.foo(foo.foo(x, y=y), y=7), y=foo.foo(123))')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('x = 101')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on stale `x`')
    run_cell('z = foo.foo(foo.foo(foo.foo(x, y=y), y=foo.foo(200)), y=99)')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('y = 102')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on stale `y`')


def test_class_member_granularity():
    run_cell("""
y = 77
class Foo:
    def __init__(self):
        self.x = 10
        self.y = y
""")
    run_cell('foo = Foo()')
    run_cell('z = foo.x + 7')
    run_cell('y = 42')
    run_cell('logging.info(z)')
    assert_not_detected('`z` independent of stale `y`')


def test_no_rhs_propagation():
    run_cell('import numpy as np')
    run_cell('x = np.random.random(10)')
    run_cell('y = np.random.random(10)')
    run_cell('inds = np.argsort(x)')
    run_cell('x = x[inds]')
    run_cell('y = y[inds]')
    assert_not_detected('`inds` should not be considered stale since it appears on RHS of assignment')


def test_if_true():
    run_cell('y = 0')
    run_cell('z = 42')
    run_cell("""
if True:
    x = y + 1
else:
    x = z + 1
""")
    run_cell('z = 43')
    run_cell('logging.info(x)')
    assert_not_detected('`x` not dependent on `z`')
    run_cell('y = 99')
    run_cell('logging.info(x)')
    assert_detected('`x` dependent on old `y`')


def test_tracing_reenabled_after_dup_funcall():
    run_cell('x = 0')
    run_cell("""
def f():
    return 42
""")
    run_cell("""
f()
f()
y = x + 1
""")
    run_cell('x = 42')
    run_cell('logging.info(y)')
    assert_detected()


def test_one_time_tracing_func():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell("""
def f(p):
    if p:
        return x
    else:
        return y
""")
    run_cell('z = f(False) + 1\nz = f(True) + 1')
    run_cell('y = 2')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('x = 3')
    run_cell('logging.info(z)')
    assert_detected('tracing should not be disabled')


def test_tracing_disable_with_nested_calls():
    # run_cell('%safety trace_messages enable')
    run_cell('y = 0')
    run_cell("""
def f():
    return y
""")
    run_cell("""
def g(flag):
    if flag:
        return f()
    else:
        return 2
""")
    run_cell("""
g(False)
x = g(True) + 1
""")
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` has dep on stale `y`')


def test_dict():
    run_cell('d = {}; d[0] = 0')
    run_cell('x = d[0] + 1')
    run_cell('d[0] = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` has dependency on old value of `d[0]`')


def test_dict_2():
    run_cell('d = {}; d[0] = 0')
    run_cell('x = d[0] + 1')
    run_cell('d = {}; d[0] = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` has dependency on old value of `d[0]`')


def test_default_dict():
    run_cell('from collections import defaultdict')
    run_cell('d = defaultdict(dict); d[0][0] = 0')
    run_cell('x = d[0][0] + 1')
    run_cell('d = defaultdict(dict); d[0][0] = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` has dependency on old value of `d[0][0]`')


def test_mutate_arg():
    run_cell('import numpy as np')
    run_cell('x = np.ones(5)')
    run_cell('y = x + 1')
    run_cell('np.random.shuffle(x)')
    run_cell('logging.info(y)')
    assert_detected('`y` has a dependency on an old value of `x`')


def test_mutate_arg_special_cases():
    run_cell('import numpy as np')
    run_cell('x = np.random.random(10)')
    run_cell('y = np.ones(5)')
    run_cell('np.random.seed(42)')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on previous seed')
    run_cell('logging.info(y)')
    assert_not_detected('`y` has no dependency on seed')
    run_cell('x = y + 5')
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell('logging.info(x)')
    assert_not_detected('`logging.info` does not mutate `y`')


def test_augassign_does_not_overwrite_deps():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell('z = x + 2')
    run_cell('z += y')
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on old value of `x`')


def test_rval_appearance_does_not_overwrite_deps():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell('z = x + 2')
    run_cell('z = z + y')
    run_cell('x = 42')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on old value of `x`')


def test_augassign_does_not_overwrite_deps_for_attributes():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x, y, z = Foo(0), Foo(5), Foo(10)')
    run_cell('z.x = x.x + 2')
    run_cell('z.x += y.x')
    run_cell('x.x = 42')
    run_cell('logging.info(z.x)')
    assert_detected('`z.x` depends on old value of `x.x`')


def test_rval_appearance_does_not_overwrite_deps_for_attributes():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('x, y, z = Foo(0), Foo(5), Foo(10)')
    run_cell('z.x = x.x + 2')
    run_cell('z.x = z.x + y.x')
    run_cell('x.x = 42')
    run_cell('logging.info(z.x)')
    assert_detected('`z.x` depends on old value of `x.x`')


def test_context_manager():
    run_cell("""
from contextlib import contextmanager

@contextmanager
def foo():
    yield 42
""")
    run_cell('with foo() as bar: x = bar + 7')
    run_cell('bar = 43')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on old value of `bar`')


def test_decorator():
    run_cell('foo = lambda f: f')
    run_cell("""
@foo
def bar(x):
    return x + 42
""")
    run_cell('y = bar(7)')
    run_cell('foo = lambda f: lambda x: f(x + 9)')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on call to `bar` which has stale decorator `@foo`')


def test_magics_dont_break_things():
    run_cell("""
%time dummy = 0
x = 0
y = x + 1
""")
    run_cell('x = 42')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on old value `x`')


def test_tuple_return():
    run_cell("""
x = 0
y = 1
a = x + 42
b = y + 77
def foo():
    return a, b
""")
    run_cell('t = foo()[1]')
    run_cell('x = 9')
    run_cell('logging.info(t)')
    assert_not_detected('`t` independent of updated `x`')
    run_cell('y = 10')
    run_cell('logging.info(t)')
    assert_detected('`t` depends on old version of `y`')


def test_tuple_return_obj():
    run_cell("""
x = 0
y = 1
a = x + 42
b = y + 77
def foo():
    return [a], [b]
""")
    run_cell('t = foo()[1][0]')
    run_cell('x = 9')
    run_cell('logging.info(t)')
    assert_not_detected('`t` independent of updated `x`')
    run_cell('y = 10')
    run_cell('logging.info(t)')
    assert_detected('`t` depends on old version of `y`')


def test_property_in_function_arg():
    run_cell("""
z = 42
class Foo:
    @property
    def bar(self):
        return z

def f(x, y):
    return x + y
""")
    run_cell('foo = Foo()')
    run_cell('w = f(3, foo.bar)')
    run_cell('z = 7')
    run_cell('logging.info(w)')
    assert_detected('`w` depends on old version of `z`')


def test_getter_setter_with_global():
    run_cell("""
z = 42
class Bar:
    shared = 25
    @property
    def baz(self):
        return z
    
    @baz.setter
    def baz(self, new_val):
        global z
        z = new_val
    
class Foo:
    def __init__(self):
        self.bar = Bar()

def f(x, y):
    return x + y
""")
    run_cell('foo = Bar()')
    run_cell('w = f(3, foo.baz)')
    run_cell('logging.info(w)')
    assert_not_detected()
    run_cell('foo.baz = 84')
    run_cell('logging.info(w)')
    assert_detected('`w` depends on stale `foo.baz`')


def test_list_extend():
    run_cell('lst = [0, 1, 2]')
    run_cell('x = lst[1] + 1')
    run_cell('lst.extend([3, 4, 5, 6, 7, 8, 9])')
    run_cell('logging.info(x)')
    assert_not_detected()
    run_cell('logging.info(lst[1])')
    assert_not_detected()
    run_cell('y = lst[8] + 1')
    run_cell('lst[8] += 42')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `lst[8]`')


@skipif_known_failing
def test_list_sum_simple():
    run_cell("w, x, y, z = 0, 1, 2, 3")
    run_cell("lst = [w, x] + [y, z]")
    run_cell("z += 42")
    run_cell("logging.info(lst[0])")
    assert_not_detected()
    run_cell("logging.info(lst[1])")
    assert_not_detected()
    run_cell("logging.info(lst[2])")
    assert_not_detected()
    run_cell("logging.info(lst[3])")
    assert_detected()


def test_global_var():
    run_cell('x = 0')
    run_cell('y = x + 1')
    run_cell('def f(): global x; x = 42')
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell('f()')
    run_cell('logging.info(y)')
    assert_detected()


def test_nonlocal_var():
    cell_template = """
z = 0
def f():
    x = 0
    def g():
        {stmt}
        x = 42
    global z
    z = x
    return g
g = f()
"""
    run_cell(cell_template.format(stmt='pass'))
    run_cell('y = z + 1')
    run_cell('g()')
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell(cell_template.format(stmt='nonlocal x'))
    run_cell('y = z + 1')
    run_cell('g()')
    run_cell('logging.info(y)')
    assert_detected()


# TODO: where was I going with this?
# def test_getitem_call():
#     run_cell("""
# class Foo:
#     def __getitem__(self, x):
#         return 42
# """)
#     run_cell('foo = Foo()')
#     run_cell('x = foo[0]')


if sys.version_info >= (3, 8):
    def test_walrus_simple():
        run_cell("""
if (x := 1) > 0:
    y = x + 1
""")
        run_cell('x = 42')
        run_cell('logging.info(y)')
        assert_detected('`y` depends on old value of `x`')

    def test_walrus_fancy():
        run_cell("""
if (x := (y := (z := 1) + 1) + 1) > 0:
    a = x + 1
""")
        run_cell('y = 42')
        run_cell('logging.info(z)')
        assert_not_detected('`z` does not depend on `y`')
        run_cell('logging.info(y)')
        assert_not_detected('`y` is updated')
        run_cell('logging.info(x)')
        assert_detected('`x` depends on old value of `y`')
        run_cell('logging.info(a)')
        assert_detected('`a` depends on old value of `y`')

    def test_walrus_fancy_attributes():
        run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
        run_cell('foo = Foo(9001)')
        run_cell("""
if (x := (y := (z := 1) + foo.x) + 1) > 0:
    a = x + 1
""")
        run_cell('foo.x = 42')
        run_cell('logging.info(z)')
        assert_not_detected('`z` does not depend on `foo.x`')
        run_cell('logging.info(y)')
        assert_detected('`y` depends on old `foo.x`')
        run_cell('logging.info(x)')
        assert_detected('`x` depends on old value of `foo.x`')
        run_cell('logging.info(a)')
        assert_detected('`a` depends on old value of `foo.x`')



# -*- coding: future_annotations -*-

# Jupyter Extension points
def _jupyter_nbextension_paths():
    return [
        dict(
            section="notebook",
            # the path is relative to the `my_fancy_module` directory
            src="resources/nbextension",
            # directory in the `nbextension/` namespace
            dest="nbsafety",
            # _also_ in the `nbextension/` namespace
            require="nbsafety/index"
        )
    ]


def load_jupyter_server_extension(nbapp):
    pass



# -*- coding: future_annotations -*-
import logging

from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def updated_symbol_names():
    return sorted([sym.readable_name for sym in nbs().updated_symbols if not sym.is_anonymous])


def run_cell(cell):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell)


def test_simplest():
    run_cell('a = 0')
    assert updated_symbol_names() == ['a']
    run_cell('a += 1')
    assert updated_symbol_names() == ['a']


def test_dict_hierarchy():
    run_cell('d = {}')
    updated_sym_names = updated_symbol_names()
    assert updated_sym_names == ['d'], 'got %s' % updated_sym_names
    run_cell('d["foo"] = {}')
    assert updated_symbol_names() == sorted(['d[foo]', 'd'])
    run_cell('d["foo"]["bar"] = []')
    updated_sym_names = updated_symbol_names()
    assert updated_sym_names == sorted(['d[foo][bar]', 'd[foo]', 'd']), 'got %s' % updated_sym_names
    run_cell('d["foo"]["bar"] = 0')
    assert updated_symbol_names() == sorted(['d[foo][bar]', 'd[foo]', 'd'])



# -*- coding: future_annotations -*-
import ast
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.extra_builtins import EMIT_EVENT, TRACING_ENABLED, make_loop_iter_flag_name
from nbsafety.singletons import nbs  # FIXME: get rid of this
from nbsafety.tracing.trace_events import TraceEvent
from nbsafety.utils import fast

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Set, Union
    from nbsafety.types import CellId


logger = logging.getLogger(__name__)


_INSERT_STMT_TEMPLATE = '{}("{{evt}}", {{stmt_id}})'.format(EMIT_EVENT)


def _get_parsed_insert_stmt(stmt: ast.stmt, evt: TraceEvent) -> ast.stmt:
    with fast.location_of(stmt):
        return fast.parse(_INSERT_STMT_TEMPLATE.format(evt=evt.value, stmt_id=id(stmt))).body[0]


def _get_parsed_append_stmt(
    stmt: ast.stmt, ret_expr: ast.expr = None, evt: TraceEvent = TraceEvent.after_stmt
) -> ast.stmt:
    with fast.location_of(stmt):
        ret = cast(ast.Expr, _get_parsed_insert_stmt(stmt, evt))
        if ret_expr is not None:
            ret_value = cast(ast.Call, ret.value)
            ret_value.keywords = fast.kwargs(ret=ret_expr)
    ret.lineno = getattr(stmt, 'end_lineno', ret.lineno)
    return ret


class StripGlobalAndNonlocalDeclarations(ast.NodeTransformer):
    def visit_Global(self, node: ast.Global) -> ast.Pass:
        with fast.location_of(node):
            return fast.Pass()

    def visit_Nonlocal(self, node: ast.Nonlocal) -> ast.Pass:
        with fast.location_of(node):
            return fast.Pass()


class StatementInserter(ast.NodeTransformer):
    def __init__(self, cell_id: Optional[CellId], orig_to_copy_mapping: Dict[int, ast.AST]):
        self._cell_id: Optional[CellId] = cell_id
        self._orig_to_copy_mapping = orig_to_copy_mapping
        self._init_stmt_inserted = False
        self._global_nonlocal_stripper = StripGlobalAndNonlocalDeclarations()

    def _handle_loop_body(self, node: Union[ast.For, ast.While], new_field: List[ast.AST]) -> List[ast.AST]:
        loop_node_copy = cast('Union[ast.For, ast.While]', self._orig_to_copy_mapping[id(node)])
        new_field.append(
            _get_parsed_append_stmt(
                cast(ast.stmt, loop_node_copy),
                evt=TraceEvent.after_loop_iter,
            )
        )
        looped_once_flag = make_loop_iter_flag_name(loop_node_copy)
        nbs().loop_iter_flag_names.add(looped_once_flag)
        with fast.location_of(loop_node_copy):
            new_field = [
                fast.If(
                    test=fast.Name(looped_once_flag, ast.Load()),
                    body=loop_node_copy.body,
                    orelse=self._global_nonlocal_stripper.visit(ast.Module(new_field)).body,
                ),
            ]
        return new_field

    def _handle_function_body(
        self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef], new_field: List[ast.AST]
    ) -> List[ast.AST]:
        fundef_copy = cast('Union[ast.FunctionDef, ast.AsyncFunctionDef]', self._orig_to_copy_mapping[id(node)])
        with fast.location_of(fundef_copy):
            new_field = [
                fast.If(
                    test=fast.parse(f'getattr(builtins, "{TRACING_ENABLED}", False)').body[0].value,  # type: ignore
                    body=new_field,
                    orelse=self._global_nonlocal_stripper.visit(fundef_copy).body,
                ),
            ]
        return new_field

    def generic_visit(self, node):
        for name, field in ast.iter_fields(node):
            if isinstance(field, ast.AST):
                setattr(node, name, self.visit(field))
            elif isinstance(field, list):
                new_field = []
                for inner_node in field:
                    if isinstance(inner_node, ast.stmt):
                        stmt_copy = cast(ast.stmt, self._orig_to_copy_mapping[id(inner_node)])
                        if not self._init_stmt_inserted:
                            self._init_stmt_inserted = True
                            with fast.location_of(stmt_copy):
                                new_field.extend(fast.parse(
                                    f'import builtins; {EMIT_EVENT}("{TraceEvent.init_cell.value}", None, cell_id="{self._cell_id}")'
                                ).body)
                        new_field.append(_get_parsed_insert_stmt(stmt_copy, TraceEvent.before_stmt))
                        if isinstance(inner_node, ast.Expr) and isinstance(node, ast.Module) and name == 'body':
                            val = inner_node.value
                            while isinstance(val, ast.Expr):
                                val = val.value
                            new_field.append(_get_parsed_append_stmt(stmt_copy, ret_expr=val))
                        else:
                            new_field.append(self.visit(inner_node))
                            if not isinstance(inner_node, ast.Return):
                                new_field.append(_get_parsed_append_stmt(stmt_copy))
                        if isinstance(node, ast.Module) and name == 'body':
                            assert not isinstance(inner_node, ast.Return)
                            new_field.append(_get_parsed_append_stmt(stmt_copy, evt=TraceEvent.after_module_stmt))
                    elif isinstance(inner_node, ast.AST):
                        new_field.append(self.visit(inner_node))
                    else:
                        new_field.append(inner_node)
                if name == 'body':
                    if isinstance(node, (ast.For, ast.While)):
                        new_field = self._handle_loop_body(node, new_field)
                    elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                        new_field = self._handle_function_body(node, new_field)
                setattr(node, name, new_field)
            else:
                continue
        return node



# -*- coding: future_annotations -*-
import ast
from typing import cast, Union, TYPE_CHECKING

from nbsafety.utils import CommonEqualityMixin

if TYPE_CHECKING:
    from typing import Any, List, Optional, Sequence, Tuple, Union
    from nbsafety.types import SupportedIndexType


class CallPoint(CommonEqualityMixin):
    def __init__(self, symbol: str):
        self.symbol = symbol

    def __hash__(self):
        return hash(self.symbol)

    def __repr__(self):
        return repr(str(self))

    def __str__(self):
        return self.symbol + '(...)'


class AttrSubSymbolChain(CommonEqualityMixin):
    def __init__(self, symbols: Sequence[Union[SupportedIndexType, CallPoint]]):
        # FIXME: each symbol should distinguish between attribute and subscript
        # FIXME: bumped in priority 2021/09/07
        self.symbols: Tuple[Union[SupportedIndexType, CallPoint], ...] = tuple(symbols)
        self.call_points = tuple(sym for sym in self.symbols if isinstance(sym, CallPoint))

    def __hash__(self):
        return hash(self.symbols)

    def __repr__(self):
        return repr(self.symbols)


class GetAttrSubSymbols(ast.NodeVisitor):
    def __init__(self):
        self.symbol_chain: List[Union[str, int, Tuple[Union[str, int], ...], CallPoint]] = []

    def __call__(self, node: Union[ast.Attribute, ast.Subscript, ast.Call, ast.Name]) -> AttrSubSymbolChain:
        self.visit(node)
        self.symbol_chain.reverse()
        return AttrSubSymbolChain(self.symbol_chain)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self.symbol_chain.append(CallPoint(node.func.attr))
            self.visit(node.func.value)
        elif isinstance(node.func, ast.Subscript):
            if isinstance(node.func.slice, ast.Constant):
                self.symbol_chain.append(CallPoint(str(node.func.slice.value)))
            elif isinstance(node.func.slice, ast.Index) and isinstance(node.func.slice.value, (ast.Str, ast.Num)):
                if isinstance(node.func.slice.value, ast.Str):
                    self.symbol_chain.append(CallPoint(node.func.slice.value.s))
                else:
                    self.symbol_chain.append(CallPoint(str(node.func.slice.value.n)))
                self.visit(node.func.value)
        elif isinstance(node.func, ast.Name):
            self.symbol_chain.append(CallPoint(node.func.id))
        elif isinstance(node.func, ast.Call):
            # TODO: handle this case too, e.g. f.g()().h
            pass
        else:
            raise TypeError('invalid type for node.func %s' % node.func)

    def visit_Attribute(self, node):
        self.symbol_chain.append(node.attr)
        self.visit(node.value)

    def visit_Subscript(self, node):
        resolved = resolve_slice_to_constant(node)
        if resolved is not None:
            if isinstance(resolved, ast.Name):
                # FIXME: hack to make the static checker stop here
                # In the future, it should try to attempt to resolve
                # the value of the ast.Name node
                self.symbol_chain.append(CallPoint(resolved.id))
            else:
                self.symbol_chain.append(resolved)
        self.visit(node.value)

    def visit_Name(self, node):
        self.symbol_chain.append(node.id)

    def generic_visit(self, node):
        # raise ValueError('we should never get here: %s' % node)
        # give up
        return


def get_attrsub_symbol_chain(maybe_node: Union[str, ast.Attribute, ast.Subscript, ast.Call]) -> AttrSubSymbolChain:
    if isinstance(maybe_node, (ast.Attribute, ast.Subscript, ast.Call)):
        node = maybe_node
    else:
        node = cast('Union[ast.Attribute, ast.Subscript, ast.Call]',
                    cast(ast.Expr, ast.parse(maybe_node).body[0]).value)
    if not isinstance(node, (ast.Attribute, ast.Subscript, ast.Call)):
        raise TypeError('invalid type for node %s' % node)
    return GetAttrSubSymbols()(node)


def resolve_slice_to_constant(node: ast.Subscript) -> Optional[Union[SupportedIndexType, ast.Name]]:
    """
    Version-independent way to get at the slice data
    """
    if isinstance(node.slice, ast.Index):
        slice = node.slice.value  # type: ignore
    else:
        slice = node.slice  # type: ignore

    if isinstance(slice, ast.Tuple):
        elts: Any = []
        for v in slice.elts:
            if isinstance(v, ast.Num):
                elts.append(v.n)
            elif isinstance(v, ast.Str):
                elts.append(v.s)
            elif isinstance(v, ast.Constant):
                elts.append(v.value)
            else:
                return None
        return tuple(elts)  # type: ignore

    negate = False
    if isinstance(slice, ast.UnaryOp) and isinstance(slice.op, ast.USub):
        negate = True
        slice = slice.operand

    if isinstance(slice, ast.Name):
        return slice

    if not isinstance(slice, (ast.Constant, ast.Str, ast.Num)):
        return None

    if isinstance(slice, ast.Constant):
        slice = slice.value
    elif isinstance(slice, ast.Num):  # pragma: no cover
        slice = slice.n  # type: ignore
        if not isinstance(slice, int):
            return None
    elif isinstance(slice, ast.Str):  # pragma: no cover
        slice = slice.s  # type: ignore
    else:
        return None

    if isinstance(slice, int) and negate:
        slice = -slice  # type: ignore
    return slice  # type: ignore



# -*- coding: future_annotations -*-
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Type


class MutationEvent:
    pass


class StandardMutation(MutationEvent):
    pass


class ListAppend(MutationEvent):
    pass


class ListExtend(MutationEvent):
    def __init__(self, orig_len: int) -> None:
        super().__init__()
        self.orig_len: int = orig_len


class ListInsert(MutationEvent):
    def __init__(self, insert_pos: Optional[int] = None):
        super().__init__()
        self.pos = insert_pos


class ListRemove(MutationEvent):
    def __init__(self, remove_pos: Optional[int] = None):
        super().__init__()
        self.pos = remove_pos


class ListPop(MutationEvent):
    def __init__(self, pop_pos: Optional[int] = None):
        super().__init__()
        self.pos = pop_pos


class NamespaceClear(MutationEvent):
    pass


class MutatingMethodEventNotYetImplemented(MutationEvent):
    pass


class ArgMutate(MutationEvent):
    pass


_METHOD_TO_EVENT_TYPE: Dict[Any, Type[MutationEvent]] = {
    list.append: ListAppend,
    list.clear: NamespaceClear,
    list.extend: ListExtend,
    list.insert: ListInsert,
    list.pop: ListPop,
    list.remove: ListRemove,
    list.sort: MutatingMethodEventNotYetImplemented,

    dict.clear: NamespaceClear,
    dict.pop: MutatingMethodEventNotYetImplemented,
    dict.popitem: MutatingMethodEventNotYetImplemented,
    dict.setdefault: MutatingMethodEventNotYetImplemented,
    dict.update: MutatingMethodEventNotYetImplemented,

    set.clear: MutatingMethodEventNotYetImplemented,
    set.difference_update: MutatingMethodEventNotYetImplemented,
    set.discard: MutatingMethodEventNotYetImplemented,
    set.intersection_update: MutatingMethodEventNotYetImplemented,
    set.pop: MutatingMethodEventNotYetImplemented,
    set.remove: MutatingMethodEventNotYetImplemented,
    set.symmetric_difference_update: MutatingMethodEventNotYetImplemented,
    set.update: MutatingMethodEventNotYetImplemented,
}


def resolve_mutating_method(obj: Any, method: Optional[str]) -> Optional[MutationEvent]:
    if method is None:
        return None
    mutation_type = _METHOD_TO_EVENT_TYPE.get(getattr(type(obj), method, None), None)
    if mutation_type is None:
        return None
    if mutation_type is ListExtend:
        return ListExtend(len(obj))
    else:
        return mutation_type()




# -*- coding: future_annotations -*-
import logging
from typing import TYPE_CHECKING

from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

if TYPE_CHECKING:
    from typing import Set

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def run_cell(cell: str, cell_id=None, **kwargs) -> None:
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, cell_id=cell_id, **kwargs)


def run_reactively(cell: str) -> Set[int]:
    executed_cells = set()
    next_cell_to_run = cell
    next_cell_to_run_id = None
    while next_cell_to_run is not None:
        run_cell(next_cell_to_run, cell_id=next_cell_to_run_id)
        next_cell_to_run = None
        fresh = sorted(nbs().check_and_link_multiple_cells().fresh_cells)
        for fresh_cell_id in fresh:
            if fresh_cell_id not in executed_cells:
                executed_cells.add(fresh_cell_id)
                next_cell_to_run = nbs().cell_content_by_cell_id[fresh_cell_id]
                next_cell_to_run_id = fresh_cell_id
                break
    return executed_cells


def test_mutate_one_list_entry():
    run_cell('lst = [1, 2, 3]')
    run_cell('logging.info(lst[0])')
    run_cell('logging.info(lst[1])')
    run_cell('logging.info(lst[2])')
    for i in range(3):
        reexeced = run_reactively(f'lst[{i}] += 1')
        assert reexeced == {i + 2}, 'got %s' % reexeced
    reexeced = run_reactively('lst.append(3)')
    assert reexeced == set(), 'got %s' % reexeced



# -*- coding: future_annotations -*-
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, List, Mapping, Optional, Sequence


class KeyDict(dict):
    def __missing__(self, key):
        return key


class GetterPipeline:
    def __init__(self, stages: Optional[List[Any]] = None):
        if stages is None:
            stages = []
        self.stages = stages

    def __ior__(self, stage):
        self.stages.append(stage)

    def __or__(self, stage):
        return GetterPipeline(self.stages + [stage])

    def __ror__(self, stage):
        return GetterPipeline([stage] + self.stages)

    def __getitem__(self, item):
        for stage in self.stages:
            item = stage[item]
        return item

    def items(self):
        for k in self.stages[0].keys():
            try:
                yield k, self[k]
            except KeyError:
                pass

    def keys(self):
        for k, _ in self.items():
            yield k

    def values(self):
        for _, v in self.items():
            yield v


class GetterFallback:
    def __init__(self, stages: Sequence[Mapping[Any, Any]]):
        self.stages = stages

    def __getitem__(self, item):
        for stage in self.stages:
            if item in stage:
                return stage[item]
        raise KeyError()

    def __setitem__(self, key, value):
        return NotImplemented

    def __contains__(self, item):
        for stage in self.stages:
            if item in stage:
                return True
        return False



# -*- coding: future_annotations -*-
import itertools
import logging
from typing import cast, TYPE_CHECKING, Sequence

from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.scope import Scope
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.singletons import nbs

if TYPE_CHECKING:
    from typing import Any, Dict, Generator, Iterable, Iterator, List, Optional, Tuple, Set
    from nbsafety.types import SupportedIndexType


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class Namespace(Scope):
    ANONYMOUS = '<anonymous_namespace>'

    PENDING_CLASS_PLACEHOLDER = object()

    # TODO: support (multiple) inheritance by allowing
    #  Namespaces from classes to clone their parent class's Namespaces
    def __init__(self, obj: Any, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.cloned_from: Optional[Namespace] = None
        self.child_clones: List[Namespace] = []
        self.obj = obj
        self.cached_obj_id = id(obj)
        if obj is not None and not isinstance(obj, int) and id(obj) in nbs().namespaces:  # pragma: no cover
            msg = 'namespace already registered for %s' % obj
            if nbs().is_develop:
                raise ValueError(msg)
            else:
                logger.warning(msg)
        if obj is not self.PENDING_CLASS_PLACEHOLDER:
            nbs().namespaces[id(obj)] = self
        self._tombstone = False
        # this timestamp needs to be bumped in DataSymbol refresh()
        self.max_descendent_timestamp: Timestamp = Timestamp.uninitialized()
        self._subscript_data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}
        self.namespace_stale_symbols: Set[DataSymbol] = set()

    @property
    def is_namespace_scope(self):
        return True

    def __bool__(self) -> bool:
        # in order to override if __len__ returns 0
        return True

    def __len__(self) -> int:
        if not isinstance(self.obj, (dict, list, tuple)):  # pragma: no cover
            raise TypeError("tried to get length of non-container namespace %s: %s", self, self.obj)
        return len(self.obj)

    def _iter_inner(self) -> Generator[Optional[DataSymbol], None, None]:
        for i in range(len(self.obj)):
            yield self.lookup_data_symbol_by_name_this_indentation(i, is_subscript=True)

    def __iter__(self) -> Iterator[Optional[DataSymbol]]:
        if not isinstance(self.obj, (list, tuple)):  # pragma: no cover
            raise TypeError("tried to iterate through non-sequence namespace %s: %s", self, self.obj)
        # do the validation before starting the generator part so that we raise immediately
        return self._iter_inner()

    def _items_inner(self) -> Generator[Tuple[Any, Optional[DataSymbol]], None, None]:
        for key in self.obj.keys():
            yield key, self.lookup_data_symbol_by_name_this_indentation(key, is_subscript=True)

    def items(self) -> Iterator[Tuple[Any, Optional[DataSymbol]]]:
        if not isinstance(self.obj, dict):  # pragma: no cover
            raise TypeError("tried to get iterate through items of non-dict namespace: %s", self.obj)
        # do the validation before starting the generator part so that we raise immediately
        return self._items_inner()

    @property
    def obj_id(self) -> int:
        return self.cached_obj_id

    @property
    def is_anonymous(self) -> bool:
        return self.scope_name == Namespace.ANONYMOUS

    @property
    def is_garbage(self) -> bool:
        return self._tombstone or self.obj_id not in nbs().aliases or self.obj_id not in nbs().namespaces

    @property
    def is_subscript(self) -> bool:
        dsym = nbs().get_first_full_symbol(self.obj_id)
        if dsym is None:
            return False
        else:
            return dsym.is_subscript

    def update_obj_ref(self, obj) -> None:
        self._tombstone = False
        nbs().namespaces.pop(self.cached_obj_id, None)
        self.obj = obj
        self.cached_obj_id = id(obj)
        nbs().namespaces[self.cached_obj_id] = self

    def data_symbol_by_name(self, is_subscript=False) -> Dict[SupportedIndexType, DataSymbol]:
        if is_subscript:
            return self._subscript_data_symbol_by_name
        else:
            return self._data_symbol_by_name

    def clone(self, obj: Any) -> Namespace:
        cloned = Namespace(obj, self.scope_name, self.parent_scope)
        cloned.cloned_from = self
        self.child_clones.append(cloned)
        return cloned

    @classmethod
    def make_child_namespace(cls, scope, scope_name) -> Namespace:
        return cls(cls.PENDING_CLASS_PLACEHOLDER, scope_name, parent_scope=scope)

    def fresh_copy(self, obj: Any) -> Namespace:
        return Namespace(obj, self.scope_name, self.parent_scope)

    def make_namespace_qualified_name(self, dsym: DataSymbol) -> str:
        path = self.full_namespace_path
        name = str(dsym.name)
        if path:
            if dsym.is_subscript:
                return f'{path}[{name}]'
            else:
                return f'{path}.{name}'
        else:
            return name

    def _lookup_subscript(self, name: SupportedIndexType) -> Optional[DataSymbol]:
        ret = self._subscript_data_symbol_by_name.get(name, None)
        if isinstance(self.obj, Sequence) and isinstance(name, int) and hasattr(self.obj, '__len__'):
            if name < 0 and ret is None:
                name = len(self.obj) + name
                ret = self._subscript_data_symbol_by_name.get(name, None)
        return ret

    def lookup_data_symbol_by_name_this_indentation(
            self,
            name: SupportedIndexType,
            *_,
            is_subscript: Optional[bool] = None,
            skip_cloned_lookup: bool = False,
            **kwargs: Any,
    ) -> Optional[DataSymbol]:
        if is_subscript is None:
            ret = self._data_symbol_by_name.get(name, None)
            if ret is None:
                ret = self._lookup_subscript(name)
        elif is_subscript:
            ret = self._lookup_subscript(name)
        else:
            ret = self._data_symbol_by_name.get(name, None)
        if not skip_cloned_lookup and ret is None and self.cloned_from is not None and not is_subscript and isinstance(name, str):
            if name not in getattr(self.obj, '__dict__', {}):
                # only fall back to the class sym if it's not present in the corresponding obj for this scope
                ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(
                    name, is_subscript=is_subscript, **kwargs
                )
        return ret

    def _remap_sym(self, from_idx: int, to_idx: int, prev_obj: Optional[Any]) -> None:
        subsym = self._subscript_data_symbol_by_name.pop(from_idx, None)
        if subsym is None:
            return
        subsym.update_usage_info()
        subsym.name = to_idx
        subsym.invalidate_cached()  # ensure we bypass equality check and bump timestamp
        subsym.update_deps(
            set(), prev_obj, overwrite=False, propagate=True, refresh=True,
        )
        self._subscript_data_symbol_by_name[to_idx] = subsym

    def shuffle_symbols_upward_from(self, pos: int) -> None:
        for idx in range(len(self.obj) - 1, pos, -1):
            prev_obj = self.obj[idx + 1] if idx < len(self.obj) - 1 else None
            self._remap_sym(idx - 1, idx, prev_obj)

    def _shuffle_symbols_downward_to(self, pos: int) -> None:
        for idx in range(pos + 1, len(self.obj) + 1):
            prev_obj = self.obj[idx - 2] if idx > pos + 1 else None
            self._remap_sym(idx, idx - 1, prev_obj)

    def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False) -> None:
        if is_subscript:
            dsym = self._subscript_data_symbol_by_name.pop(name, None)
            if dsym is None and name == -1 and isinstance(self.obj, list):
                name = len(self.obj)  # it will have already been deleted, so don't subtract 1
                dsym = self._subscript_data_symbol_by_name.pop(name, None)
            if dsym is not None:
                dsym.update_deps(set(), deleted=True)
            if isinstance(self.obj, list) and isinstance(name, int):
                self._shuffle_symbols_downward_to(name)
        else:
            super().delete_data_symbol_for_name(name)

    def all_data_symbols_this_indentation(self, exclude_class=False, is_subscript=None) -> Iterable[DataSymbol]:
        if is_subscript is None:
            dsym_collections_to_chain: List[Iterable] = [
                self._data_symbol_by_name.values(), self._subscript_data_symbol_by_name.values()
            ]
        elif is_subscript:
            dsym_collections_to_chain = [self._subscript_data_symbol_by_name.values()]
        else:
            dsym_collections_to_chain = [self._data_symbol_by_name.values()]
        if self.cloned_from is not None and not exclude_class:
            dsym_collections_to_chain.append(self.cloned_from.all_data_symbols_this_indentation())
        return itertools.chain(*dsym_collections_to_chain)

    def put(self, name: SupportedIndexType, val: DataSymbol) -> None:
        if val.is_subscript:
            self._subscript_data_symbol_by_name[name] = val
        elif not isinstance(name, str):  # pragma: no cover
            raise TypeError('%s should be a string' % name)
        else:
            self._data_symbol_by_name[name] = val
        val.containing_scope = self

    def refresh(self) -> None:
        self.max_descendent_timestamp = Timestamp.current()

    def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> Optional[Namespace]:
        # TODO: test this properly
        ret = None
        if self.namespace_parent_scope is not None:
            ret = self.namespace_parent_scope.get_earliest_ancestor_containing(obj_id, is_subscript)
        if ret is not None:
            return ret
        if obj_id in (dsym.obj_id for dsym in self.all_data_symbols_this_indentation(is_subscript=is_subscript)):
            return self
        else:
            return None

    @property
    def namespace_parent_scope(self) -> Optional[Namespace]:
        if self.parent_scope is not None and isinstance(self.parent_scope, Namespace):
            return self.parent_scope
        return None

    def transfer_symbols_to(self, new_ns: Namespace) -> None:
        for dsym in list(self.all_data_symbols_this_indentation(exclude_class=True, is_subscript=False)):
            try:
                inner_obj = nbs().retrieve_namespace_attr_or_sub(new_ns.obj, dsym.name, is_subscript=False)
            except AttributeError:
                inner_obj = None
            dsym.update_obj_ref(inner_obj)
            logger.info("shuffle %s from %s to %s", dsym, self, new_ns)
            self._data_symbol_by_name.pop(dsym.name, None)
            new_ns._data_symbol_by_name[dsym.name] = dsym
            dsym.containing_scope = new_ns
        for dsym in list(self.all_data_symbols_this_indentation(exclude_class=True, is_subscript=True)):
            try:
                inner_obj = nbs().retrieve_namespace_attr_or_sub(new_ns.obj, dsym.name, is_subscript=True)
            except (IndexError, KeyError):
                inner_obj = None
            dsym.update_obj_ref(inner_obj)
            logger.info("shuffle %s from %s to %s", dsym, self, new_ns)
            self._subscript_data_symbol_by_name.pop(dsym.name, None)
            new_ns._subscript_data_symbol_by_name[dsym.name] = dsym
            dsym.containing_scope = new_ns



# -*- coding: future_annotations -*-
import logging

from .utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def test_stack_on_tracing_reenable():
    run_cell("""
x = 42

def fake_func(y):
    z = 9
    return z

new_xs = [fake_func(x) for _ in range(5)]

logging.info(new_xs[0])
""")


def test_non_idempotent_subscript():
    run_cell("""
class IncDict:
    def __init__(self):
        self.val = 0
    def __getitem__(self, val):
        self.val += val
        return self.val
        
d = IncDict()

d_sub_1 = d[1]
assert d_sub_1 == 1, f"got {d_sub_1}, but expected 1"
""")


def test_starred_args():
    run_cell("""
def f(foo, bar):
    return foo + bar
""")
    run_cell('args = [1, 2]')
    run_cell('f(*args)')


def test_starred_assignment():
    run_cell('x = 0')
    run_cell('y = 1')
    run_cell('z = 2')
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell('s, *t = [x + 1, y + 2, z + 3, *lst]')
    run_cell('z = 42')
    run_cell('logging.info(s)')
    run_cell('logging.info(t[0])')
    run_cell('logging.info(t[1])')
    run_cell('x = 99')
    run_cell('logging.info(s)')
    run_cell('logging.info(t[0])')


def test_slices():
    run_cell('lst = list(range(10))')
    run_cell('foo = lst[3:7]')
    run_cell('lst[1:2] = foo')


def test_partial_slices():
    run_cell('s = "Hello, world!"')
    run_cell('logging.info(str(s[:7]))')
    run_cell('logging.info(str(s[7:]))')


def test_delete():
    run_cell('lst = list(range(10))')
    run_cell('del lst[0]')
    run_cell('del lst[-1]')


def test_delitem():
    run_cell("""
class Foo:
    def __init__(self):
        self.lst = []
    def __getitem__(self, i):
        return self.lst[i]
    def __setitem__(self, i, v):
        self.lst[i] = v
    def __delitem__(self, i):
        del self.lst[i]
""")
    run_cell('foo = Foo()')
    run_cell('foo.lst.append(0)')
    run_cell('foo.lst.append(1)')
    run_cell('foo.lst.append(2)')
    run_cell('logging.info(foo[0])')
    run_cell('foo[1] = 2')
    run_cell('del foo[-1]')


def test_empty_return():
    run_cell("""
def foo():
    return
""")
    run_cell('x = foo()')


def test_fancy_slices():
    run_cell("""
class Foo:
    def __init__(self, x):
        self.x = x
""")
    run_cell('foo = Foo(1)')
    run_cell('import numpy as np')
    run_cell('x = np.zeros((3, 3, 3))')
    run_cell('logging.info(x[:,:,:])')
    run_cell('logging.info(x[:,...])')
    run_cell('logging.info(x[1:,...])')
    run_cell('logging.info(x[foo.x])')
    run_cell('logging.info(x[foo.x:,...])')
    run_cell('logging.info(x[foo.x:foo.x+1,:,...])')


def test_fancy_slice_assign_augassign():
    run_cell('%safety trace_messages enable')
    run_cell('import numpy as np')
    run_cell('x = np.zeros((3, 3, 3))')
    run_cell('x[:, 1 ,...] = 1.')
    run_cell('x[:, 1 ,...] /= 5.')


def test_pass():
    run_cell('if True: pass')


def test_none_key():
    run_cell('d = {}')
    run_cell('d[None] = None')
    run_cell('d.clear()')


def test_global_var():
    run_cell('x = 0')
    run_cell('def f(): global x; x = 42')
    run_cell('f()')
    run_cell('assert x == 42')
